local _DEBUG = true
local _UPDATEDISPLAY_RATE = 0.2
local _SCANTALENT_RATE = 1
local SPELL_SWP = GetSpellInfo(589)
local CR_HASTE_SPELL = CR_HASTE_SPELL or 20
local DPS_COMPUTATION_WINDOW_LENGTH = 30

local COMBATLOG_OBJECT_AFFILIATION_MINE		= COMBATLOG_OBJECT_AFFILIATION_MINE		or 0x00000001
local COMBATLOG_OBJECT_REACTION_HOSTILE		= COMBATLOG_OBJECT_REACTION_HOSTILE		or 0x00000040
local COMBATLOG_OBJECT_TYPE_NPC				= COMBATLOG_OBJECT_TYPE_NPC				or 0x00000800
local hostileNPCMask = COMBATLOG_OBJECT_REACTION_HOSTILE+COMBATLOG_OBJECT_TYPE_NPC

ShadowGreenLight = LibStub("AceAddon-3.0"):NewAddon("ShadowGreenLight", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")
local TalentQuery = LibStub:GetLibrary("LibTalentQuery-1.0")

-- Option table
local options = { 
    name = "ShadowGreenLight",
    handler = ShadowGreenLight,
    type = "group",
    args = {
        displayOptions = {
        	name = "Display options",
		    handler = ShadowGreenLight,
		    type = "group",
		    args = {
		    },
		},
        selectDisplay = {
        	type = "input",
			name = "Display plugin selection",
			desc = "Select display plugin to use",
			get = function(info)
				return ShadowGreenLight.db.profile.selectedDisplay
			end,
			set = function(info, v)
				ShadowGreenLight.db.profile.selectedDisplay = v
			end,
			validate = function(info)
				return ShadowGreenLight.registeredDisplays
			end,
		},
        help = {
            type = "execute",
            name = "Help",
            desc = "Display a description of the different indicators",
            func = "OutputHelp",
        },
    },
}

function ShadowGreenLight:OnInitialize()
  	-- Initialization of attributes
  	self.maxUpgrade = 0
	self.currentUpgrade = 0
	self.lastSWPUpgrade = 0
	self.critPctWithoutRaidBuffs = 0
	self.upgradeList = {}
	self.currentTooltipUpgradeList = {}
	self.raidUpgradeAvailable = {}
	self.scanStatus = {}
	self.isInCombat = false
	self.dpslog = {
		["first"] = 1,
		["last"] = 0,
	}
	self.lastSWPTick = 0
	
	-- Initialization of spell database
	self.upgradeDb = {
  		["ShadowWeave"] = {
  			name = GetSpellInfo(15257),
			maxCount = 5,
			upgradePerStack = 2,
			type = "Shadow Weaving",
			detectionUnit = "player",
			detectionType = "buff",
			upgradeType = "dmg",
		},
		["Improved Scorch"] = {
  			name = GetSpellInfo(11095),
  			talentName = GetSpellInfo(11095),
  			talentTab = 2,
  			talentIndex = 11,
			maxCount = 5,
			upgradePerStack = 2,
			type = "Spell Critical Strike Chance Debuff",
			detectionUnit = "target",
			detectionType = "debuff",
			upgradeType = "crit",
		},
		["Winter's Chill"] = {
  			name = GetSpellInfo(11180),
  			talentName = GetSpellInfo(11180),
  			talentTab = 3,
  			talentIndex = 18,
			maxCount = 5,
			upgradePerStack = 2,
			type = "Spell Critical Strike Chance Debuff",
			detectionUnit = "target",
			detectionType = "debuff",
			upgradeType = "crit",
		},
		["Totem of Wrath"] = {
  			name = GetSpellInfo(57722),
  			talentName = GetSpellInfo(57722),
  			talentTab = 1,
  			talentIndex = 21,
			maxCount = 1,
			upgradePerStack = 3,
			type = "Critical Strike Chance Taken Debuff",
			detectionUnit = "target",
			detectionType = "debuff",
			upgradeType = "crit",
		},
		["Heart of the Crusader"] = {
  			name = GetSpellInfo(20337),
  			talentName = GetSpellInfo(20337),
  			talentTab = 3,
  			talentIndex = 4,
			maxCount = 1,
			upgradePerTalentRank = {
				[1] = 1,
				[2] = 2,
				[3] = 3,
			},
			type = "Critical Strike Chance Taken Debuff",
			detectionUnit = "target",
			detectionType = "debuff",
			upgradeType = "crit",
		},
		["Moonkin Aura"] = {
  			name = GetSpellInfo(24907),
  			talentName = GetSpellInfo(24858),
  			talentTab = 1,
  			talentIndex = 18,
			maxCount = 1,
			upgradePerStack = 5,
			type = "Spell Critical Strike Chance Buff",
			detectionUnit = "player",
			detectionType = "buff",
			upgradeType = "crit",
		},
		["Elemental Oath"] = {
  			name = GetSpellInfo(53414),
  			talentName = GetSpellInfo(53414),
  			talentTab = 1,
  			talentIndex = 18,
			maxCount = 1,
			upgradePerTalentRank = {
				[1] = 3,
				[2] = 5,
			},
			type = "Spell Critical Strike Chance Buff",
			detectionUnit = "player",
			detectionType = "buff",
			upgradeType = "crit",
		},
		["Tricks of the Trade"] = {
  			name = GetSpellInfo(57934),
			upgradePerStack = 6,
			type = "Tricks of the Trade",
			detectionUnit = "player",
			detectionType = "buff",
			isBonus = true,
			upgradeType = "dmg",
		},
		["Ferocious Inspiration"] = {
  			name = GetSpellInfo(34455),
  			talentName = GetSpellInfo(34455),
  			talentTab = 1,
  			talentIndex = 17,
			maxCount = 1,
			upgradePerTalentRank = {
				[1] = 1,
				[2] = 2,
				[3] = 3,
			},
			type = "Percentage Damage Increase",
			detectionUnit = "player",
			detectionType = "buff",
			upgradeType = "dmg",
		},
		["Sanctified Retribution"] = {
  			name = GetSpellInfo(7294),
  			talentName = GetSpellInfo(31869),
  			talentTab = 3,
  			talentIndex = 14,
			maxCount = 1,
			upgradePerStack = 3,
			type = "Percentage Damage Increase",
			detectionUnit = "player",
			detectionType = "buff",
			upgradeType = "dmg",
		},
		["Power Spark"] = {
  			name = GetSpellInfo(56152),
			upgradePerStack = 50,
			type = "Encounter Specific",
			detectionUnit = "player",
			detectionType = "debuff",
			isBonus = true,
			upgradeType = "dmg",
		},
		["Fungal Creep"] = {
  			name = GetSpellInfo(29232),
			upgradePerStack = 50,
			type = "Encounter Specific",
			detectionUnit = "player",
			detectionType = "debuff",
			isBonus = true,
			upgradeType = "crit",
		},
		["Negative Charge"] = {
  			name = GetSpellInfo(39091),
			upgradePerStack = 10,
			type = "Encounter Specific",
			detectionUnit = "player",
			detectionType = "debuff",
			isBonus = true,
			upgradeType = "dmg",
		},
		["Positive Charge"] = {
  			name = GetSpellInfo(39091),
			upgradePerStack = 10,
			type = "Encounter Specific",
			detectionUnit = "player",
			detectionType = "debuff",
			isBonus = true,
			upgradeType = "dmg",
		},
  	}
  	
  	-- Initialization of session db
  	self.db = LibStub("AceDB-3.0"):New("ShadowGreenLightDB", {
  		profile = {
  			selectedDisplay = "Default",
		}
	}, "Default")
	
	-- Adding display options for registered display plugins
	self.options = options
	for name, plugin in pairs(self.registeredDisplays) do
		self.options.args.displayOptions.args[name] = {
			["name"] = name,
			["handler"] = object,
			["type"] = "group",
			["args"] = plugin.options,
		}
	end
	
	-- Registering chat commands
	LibStub("AceConfig-3.0"):RegisterOptionsTable("ShadowGreenLight", self.options, {"ShadowGreenLight", "sgl"})
end

function ShadowGreenLight:OnEnable()
	-- Registering player entering world
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "RegisterPertinentEvents")
end

function ShadowGreenLight:OnDisable()
    -- Callback for talent query
    TalentQuery.UnregisterCallback(self, "TalentQuery_Ready")
end

function ShadowGreenLight:Debug(...)
	if _DEBUG then
		self:Print(...)
	end
end

function ShadowGreenLight:RegisterPertinentEvents()
	local _, class = UnitClass("player")
	
	if class == "PRIEST" then
		-- Registering event for group / raid joining
		self:RegisterEvent("PARTY_MEMBERS_CHANGED", "RosterUpdate")
		-- Registering events to enter/exit combat
	    self:RegisterEvent("PLAYER_REGEN_DISABLED", "EnteringCombat")
	    self:RegisterEvent("PLAYER_REGEN_ENABLED", "ExitingCombat")
	    -- Callback for talent query
	    TalentQuery.RegisterCallback(self, "TalentQuery_Ready")
	    -- Create display
	    self:CreateDisplay()
	end
end



--------------------------------------------------
-- Display registering methods
function ShadowGreenLight:RegisterDisplay(name, object)
	if not self.registeredDisplays then
		self.registeredDisplays = {}
	end
	self.registeredDisplays[name] = object
end

function ShadowGreenLight:OutputHelp()
	self:Print("ShadowGreenLight aims at helping shadow priests decide whether refreshing SWP is profitable, based on which buffs / debuffs are available in your raid")
	self:Print("It uses different indicators to achieve this goal :")
	self:Print("- Background color of the frame : from red (no available upgrades are up) to green (all available upgrades are up)")
	self:Print("- Numeric indicator (top of the frame) : same information as background color, except displayed as numeric value in terms of DPS increase (current DPS increase / max available DPS increase)")
	self:Print("- Skull in the frame : recasting SW:P now would result in an increase of its tick values, however keep in mind that you would be using a gcd to refresh SW:P and thus might actually lose DPS, next indicator might give you a better idea.")
	self:Print("- Time to profit indicator (bottom of the frame) : according to your current DPS and the upgrades available, this indicator tells you how long it will take to be compensated from losing a gcd to refresh SWP")
	self:Print("- Tooltip : hovering the frame in combat will display a recap of which upgrades are currently available, time to scream at your mage to have better scorch uptime !")
end

--------------------------------------------------
-- Display methods
function ShadowGreenLight:CreateDisplay()
	local display = self.registeredDisplays[self.db.profile.selectedDisplay]
	if not display then display = self.registeredDisplays["Default"] end
	
	display:CreateDisplay()
end

function ShadowGreenLight:UpdateDisplay()
	local display = self.registeredDisplays[self.db.profile.selectedDisplay]
	if not display then display = self.registeredDisplays["Default"] end
	
	display:UpdateDisplay()
end

--------------------------------------------------
-- Internal logic methods

function ShadowGreenLight:RosterUpdate()
	-- Schedule timer to scan talents if in raid or group
	if (GetNumRaidMembers() > 0 or GetNumPartyMembers() > 0) then
		if not self.talentScanTimer then
			self.talentScanTimer = self:ScheduleRepeatingTimer("ScanRaidTalents", _SCANTALENT_RATE)
			self:Debug("Scheduling timer for talent scan")
		end
	-- Unschedule timer to scan talents if not in a raid or group
	elseif self.talentScanTimer then
		self:CancelTimer(self.talentScanTimer)
		self.talentScanTimer = nil
		self:Debug("Unscheduling timer for talent scan")
	end
end

function ShadowGreenLight:EnteringCombat()
	local display = self.registeredDisplays[self.db.profile.selectedDisplay]
	if not display then display = self.registeredDisplays["Default"] end
	
	self:RegisterEvent("PLAYER_TARGET_CHANGED", "UpdateDisplay")
	self:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED", "handleSpellcast")
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED", "handleCombatLogEvent")
	self.updateDisplayTimer = self:ScheduleRepeatingTimer("UpdateDisplay", _UPDATEDISPLAY_RATE)
	
	display:UpdateDisplay_OnEnteringCombat()
		
	self.isInCombat = true
	
	-- Resetting dpslog
	local log = self.dpslog
	for i = log.first, log.last do
		log[i] = nil
	end
	log.first = 1
	log.last = 0
	self.lastSWPTick = 0
	
	self:PopulateUpgradeList()
	self:ComputeMaxUpgrade()
	self:ComputeCritPctWithoutRaidBuffs()
end

function ShadowGreenLight:ExitingCombat()
	self:UnregisterEvent("PLAYER_TARGET_CHANGED")
	self:UnregisterEvent("UNIT_SPELLCAST_SUCCEEDED")
	self:CancelTimer(self.updateDisplayTimer)
	self.updateDisplayTimer = nil
	self.lastSWPUpgrade = 0
	
	local display = self.registeredDisplays[self.db.profile.selectedDisplay]
	if not display then display = self.registeredDisplays["Default"] end
	
	display:UpdateDisplay_OnExitingCombat()
	
	self.isInCombat = false
end

--------------------------------------------------
-- Detecting SW:P application
function ShadowGreenLight:handleSpellcast(event, unit, spellName, ...)
    if unit == "player" and spellName == SPELL_SWP then
        local currentUpgrade = self:ComputeCurrentUpgrade()
        self.lastSWPUpgrade = currentUpgrade 
    end
end

--------------------------------------------------
-- Logging dmg events
function ShadowGreenLight:handleCombatLogEvent(event, timeStamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
	
    if srcName == UnitName("player") and dstName == UnitName("target") then
		local prefix, suffix, special = strsplit("_", eventType)
		if suffix == "DAMAGE" or special=="DAMAGE" then
			local _, spellName, _, dmg = ...
	    	local log = self.dpslog
	    	-- logging new entry
			local entry = {}
			entry.timeStamp = timeStamp
			entry.dmg = dmg
			log.last = log.last + 1
			log[log.last] = entry
			-- cleanup old entries
			for i = log.first, log.last do
				if log[i].timeStamp < timeStamp - DPS_COMPUTATION_WINDOW_LENGTH then
					log[i] = nil
					log.first = log.first + 1
				else
					break
				end
			end
			-- log last SWP tick value
			if spellName == SPELL_SWP then
				self.lastSWPTick = dmg
			end
	    end
	end
end

---------------------------------------------------
-- Talent scanning

function ShadowGreenLight:ScanRaidTalents()
	if GetNumRaidMembers()>0 then
		for i = 1,40 do
			self:ScanUnitIfNotReceived("raid"..i)
		end
	else
		if GetNumPartyMembers() > 0 then
			for i = 1,4 do
				self:ScanUnitIfNotReceived("party"..i)
			end
		end
	end
end

function ShadowGreenLight:ScanUnitIfNotReceived(unit)
	if UnitExists(unit) then
		if not self.scanStatus[UnitName(unit)] or self.scanStatus[UnitName(unit)] ~= "RECEIVED" then
			self.scanStatus[UnitName(unit)] = "SENT"
			if UnitName(unit)==UnitName("player") then
				self:TalentQuery_Ready(_, UnitName(unit))
			else
				TalentQuery:Query(unit)
			end
			return true
		end
	end
	return false
end

-- Handle LibTalentQuery-1.0 TalentQuery_Ready event
function ShadowGreenLight:TalentQuery_Ready(_, name)
	self:Debug("Talent info received for " .. name)
	local isNotPlayer = (UnitName("player")~=name)
	-- Scan spells in upgrade database (except shadowweave)
	for upgradeName, upgrade in pairs(self.upgradeDb) do
		-- skip non talents
		if upgrade.talentName then
			-- Get talent related to upgrade 
			local talentName, _, _, _, talentRank = GetTalentInfo(upgrade.talentTab, upgrade.talentIndex, isNotPlayer)
			if talentName and talentName == upgrade.talentName then
				if talentRank > 0 then
					if not self.raidUpgradeAvailable[name] then
						self.raidUpgradeAvailable[name] = {}
					end
					if not self.raidUpgradeAvailable[name][upgradeName] then
						self.raidUpgradeAvailable[name][upgradeName] = {
							name = upgrade.name,
				  			maxCount = upgrade.maxCount,
							upgradePerStack = upgrade.upgradePerStack,
							type = upgrade.type,
							detectionUnit = upgrade.detectionUnit,
							detectionType = upgrade.detectionType,
							upgradeType = upgrade.upgradeType,
						}
						-- upgrade per stack can be dependant on talent point investment
						if upgrade.upgradePerTalentRank then
							self.raidUpgradeAvailable[name][upgradeName].upgradePerStack = upgrade.upgradePerTalentRank[talentRank]
						end
					end
					self:Debug(name .. " > " .. upgradeName .. " > " .. self.raidUpgradeAvailable[name][upgradeName].upgradePerStack)
				end 
			end
		end
	end
	self.scanStatus[name] = "RECEIVED"
end

function ShadowGreenLight:ComputeCurrentUpgrade()
	local currentUpgradeList = {
		["crit"]={},
		["dmg"]={},
	}
	
	-- Reset current stack count in upgrade list for tooltip
	for upgradeType in pairs(self.currentTooltipUpgradeList) do
	    for upgradeName in pairs(self.currentTooltipUpgradeList[upgradeType]) do
	       self.currentTooltipUpgradeList[upgradeType][upgradeName].currentStack = 0
	    end
	end
	
	-- Scan upgrade list
	for _, upgradeBuff in pairs(self.upgradeList) do
		-- detection function is either UnitBuff or UnitDebuff according to detectionType
		local detectionFunction
		if upgradeBuff.detectionType == "buff" then
			detectionFunction = UnitBuff
		else
			detectionFunction = UnitDebuff
		end
		
		-- Cycle through every buffs or debuffs of detection unit
		local index = 1
		local auraName, _, _, auraCount = detectionFunction(upgradeBuff.detectionUnit, index)
		while auraName do
			if upgradeBuff.name == auraName then
			    if auraCount == 0 then auraCount = 1 end -- apparently if buff does not stack UnitBuff returns 0
				if not currentUpgradeList[upgradeBuff.upgradeType][upgradeBuff.type] then
					currentUpgradeList[upgradeBuff.upgradeType][upgradeBuff.type] = 0
				end				
				local upgradeAmount = upgradeBuff.upgradePerStack * auraCount
				-- Check if need updating
				if upgradeAmount > currentUpgradeList[upgradeBuff.upgradeType][upgradeBuff.type] then
					currentUpgradeList[upgradeBuff.upgradeType][upgradeBuff.type] = upgradeAmount
				end
				-- Update tooltip
				self.currentTooltipUpgradeList[upgradeBuff.type][upgradeBuff.name].currentStack = auraCount
			end
			index = index + 1
			auraName, _, _, auraCount = detectionFunction(upgradeBuff.detectionUnit, index)
		end
	end
	
	local currentCritFactor = 0
	for _, upgradeAmount in pairs(currentUpgradeList["crit"]) do
		currentCritFactor = currentCritFactor + upgradeAmount
	end
	
	local currentDmgFactor = 1
	for _, upgradeAmount in pairs(currentUpgradeList["dmg"]) do
		currentDmgFactor = currentDmgFactor * (1 + upgradeAmount/100)
	end
	
	local currentUpgrade = ((1+(1+self.critPctWithoutRaidBuffs/100)*currentCritFactor) * currentDmgFactor - 1)*100
	return currentUpgrade
end

---------------------------------------------------
-- Populate Upgrade List
function ShadowGreenLight:PopulateUpgradeList()
	-- reinitialize upgrade list
	for upgradeName, _ in pairs(self.upgradeList) do
		self.upgradeList[upgradeName] = nil
	end
	
	-- Always assume user is a shadowpriest with maxed shadow weaving (duh ?)
	self.upgradeList["ShadowWeave"] = self.upgradeDb["ShadowWeave"]
	
	-- taking list of raid member and comparing it to available upgrade list
	-- skip afk / not visible / out of zone players
	if GetNumRaidMembers()>0 then
		for i = 1,40 do
		    if UnitIsVisible("raid"..i) then
        		self:PopulateUpgradeListForUnit("raid"..i)
        	end
		end
	else
		if GetNumPartyMembers() > 0 then
			for i = 1,4 do
			    if UnitIsVisible("party"..i) then
        		  self:PopulateUpgradeListForUnit("party"..i)
        	    end
			end
		end
	end
    
    -- Reset current upgrade list for tooltip
	for upgradeType in pairs(self.currentTooltipUpgradeList) do
	    for upgradeName in pairs(self.currentTooltipUpgradeList[upgradeType]) do
	       self.currentTooltipUpgradeList[upgradeType][upgradeName].currentStack = nil
	       self.currentTooltipUpgradeList[upgradeType][upgradeName].maxStack = nil
	       self.currentTooltipUpgradeList[upgradeType][upgradeName] = nil
	    end
		self.currentTooltipUpgradeList[upgradeType] = nil
	end
    -- register upgrades for tooltip
	for _, upgradeBuff in pairs(self.upgradeList) do
        if not self.currentTooltipUpgradeList[upgradeBuff.type] then
	        self.currentTooltipUpgradeList[upgradeBuff.type] = {}
        end
        -- register upgrade for tooltip
        self.currentTooltipUpgradeList[upgradeBuff.type][upgradeBuff.name] = {
    	     currentStack = 0,
    	     maxStack = upgradeBuff.maxCount
        } 
    end
end

function ShadowGreenLight:PopulateUpgradeListForUnit(unit)
	local name = UnitName(unit)
	if self.raidUpgradeAvailable[name] then
		for upgradeName, upgrade in pairs(self.raidUpgradeAvailable[name]) do
			-- either buff was not available before
			if not self.upgradeList[upgradeName] then
				self.upgradeList[upgradeName] = upgrade
			-- or it was, in that case check if new upgrade is better
			else
				if self.upgradeList[upgradeName].upgradePerStack < upgrade.upgradePerStack then
					self.upgradeList[upgradeName] = upgrade
				end
			end
		end
	end
end

function ShadowGreenLight:ComputeMaxUpgrade()
	local maxUpgradeByUpgradeType = {
		["crit"]={},
		["dmg"]={},
	}
	self.maxUpgrade = 0
	for _, upgradeBuff in pairs(self.upgradeList) do
		if not upgradeBuff.isBonus then	-- skip bonus buffs for maxupgrade computation
			if not maxUpgradeByUpgradeType[upgradeBuff.upgradeType][upgradeBuff.type] then
				maxUpgradeByUpgradeType[upgradeBuff.upgradeType][upgradeBuff.type] = 0
			end
			local maxUpgradeAmount = upgradeBuff.upgradePerStack * upgradeBuff.maxCount
			if maxUpgradeAmount > maxUpgradeByUpgradeType[upgradeBuff.upgradeType][upgradeBuff.type] then
				maxUpgradeByUpgradeType[upgradeBuff.upgradeType][upgradeBuff.type] = maxUpgradeAmount
			end
		end
	end
	
	local maxCritFactor = 0
	for _, upgradeAmount in pairs(maxUpgradeByUpgradeType["crit"]) do
		maxCritFactor = maxCritFactor + upgradeAmount
	end
	
	local maxDmgFactor = 1
	for _, upgradeAmount in pairs(maxUpgradeByUpgradeType["dmg"]) do
		maxDmgFactor = maxDmgFactor * (1 + upgradeAmount/100)
	end
	
	self.maxUpgrade = ((1+(1+self.critPctWithoutRaidBuffs/100)*maxCritFactor) * maxDmgFactor - 1)*100
	return self.maxUpgrade
end

function ShadowGreenLight:ComputeCritPctWithoutRaidBuffs()
	local maxCritUpgradeByUpgradeType = {}
	self.critPctWithoutRaidBuffs = GetSpellCritChance(6)
		
	-- Scan upgrade list
	for _, upgradeBuff in pairs(self.upgradeList) do
		-- only keep self crit buffs
		if upgradeBuff.detectionType == "buff" and upgradeBuff.detectionUnit == "player" and upgradeBuff.upgradeType == "crit" then
			-- Cycle through every buffs or debuffs of detection unit
			local index = 1
			local auraName, _, _, auraCount = UnitBuff("player", index)
			while auraName do
				if upgradeBuff.name == auraName then
				    if auraCount == 0 then auraCount = 1 end -- apparently if buff does not stack UnitBuff returns 0
					if not maxCritUpgradeByUpgradeType[upgradeBuff.type] then
						maxCritUpgradeByUpgradeType[upgradeBuff.type] = 0
					end				
					local upgradeAmount = upgradeBuff.upgradePerStack * auraCount
					-- Check if need updating
					if upgradeAmount > maxCritUpgradeByUpgradeType[upgradeBuff.type] then
						maxCritUpgradeByUpgradeType[upgradeBuff.type] = upgradeAmount
					end
				end
				index = index + 1
				auraName, _, _, auraCount = UnitBuff("player", index)
			end
		end
	end
	
	for _, upgradeAmount in pairs(maxCritUpgradeByUpgradeType) do
		self.critPctWithoutRaidBuffs = self.critPctWithoutRaidBuffs - upgradeAmount
	end
	
	return self.critPctWithoutRaidBuffs
end

function ShadowGreenLight:GetTimeToProfit(currentUpgrade)
	-- Compute dps (0 if not enough data available)
	local totalLoggedDmg = 0
	local log = self.dpslog
	-- If at least 10 entries in the DPS computation window
	if log.last - log.first >= 10 then
		for i = log.first, log.last do
			totalLoggedDmg = totalLoggedDmg + log[i].dmg
		end
	end
	
	local dps = 0
	if log[log.last] and log[log.first] then
		dps = totalLoggedDmg / (log[log.last].timeStamp - log[log.first].timeStamp)
	end
	
	-- Compute gcd value with haste
	local hasteRating = GetCombatRating(CR_HASTE_SPELL)
	local gcd = 1.5 / (1 + (hasteRating / 32.79 / 100))
	if gcd < 1 then gcd = 1 end
	
	-- Compute SWP DPS
	local spwDps = self.lastSWPTick/3
	
	local ttp = gcd * dps / (spwDps * (currentUpgrade - self.lastSWPUpgrade)/100)
	
	if dps == 0 then
		return "--"
	else
		return string.format("%.0f'%.0f\"", math.floor(ttp / 60), ttp % 60)
	end
end